#!/usr/bin/env python3

import itertools
import enum

import numpy as np
import scipy.spatial

import lapjv

from .common import ConfigurationError
from .tools import LoggingObserver, warning, shortest_warp_path
from .config import CONFIG
from .dtw import dynamic_timewarping


class AlgoState(enum.Enum):
	MATCHCOST_UPDATED = 1
	WARPCOST_UPDATED = 2
	CONVERGED = 3
	ITERATION_LIMIT_REACHED = 4


class DtgwHeuristic:
	def __init__(self, features1, features2, metric):
		t1, n, k = features1.shape
		t2, n2, k2 = features2.shape
		if k != k2 or n != n2:
			raise ValueError("Arguments have wrong dimensions")

		self._features1 = features1
		self._features2 = features2
		self._k = k
		self._n = n
		self._t1 = t1
		self._t2 = t2
		self._metric = metric

		self._warppath = []
		self._matching = []
		self._warpcost = None
		self._matchcost = None
		self._state = None
		self._cost = float("inf")
		self.iteration = 0
		self.max_iterations = float("inf")
		# minimal number of iterations required to obtain a valid result:
		self._min_iterations = None

	@property
	def terminated(self):
		return self._state in (AlgoState.CONVERGED, AlgoState.ITERATION_LIMIT_REACHED)

	@property
	def warp_path(self):
		return self._warppath
	
	@property
	def matching(self):
		return self._matching

	@property
	def cost(self):
		return self._cost

	@property
	def state(self):
		return self._state

		
	def initialize_with_diagonal_warping(self):
		self._warppath = list(shortest_warp_path(self._t1, self._t2))
		self._update_matchcost()
		self._min_iterations = 0

	def initialize_with_optimistic_warping(self):
		self._warpcost = np.empty((self._t1, self._t2))
		for i, j in np.ndindex(self._t1, self._t2):
			self._warpcost[i,j] = lapjv.lapjv(self._metric(self._features1[i,:], self._features2[j,:]))[2][0]
		self._state = AlgoState.WARPCOST_UPDATED
		self._min_iterations = 2

	def initiialize_with_product_warping(self):
		self._warppath = list(itertools.product(range(self._t1), range(self._t2)))
		self._update_matchcost()
		self._min_iterations = 0

	def initialize_with_optimistic_matching(self):
		self._matchcost = np.zeros((self._n, self._n))
		for i in range(self._t1):
			tmp = np.stack(
				(
					self._metric(self._features1[i,:,:], self._features2[j,:,:])
					for j in range(self._t2)
				),
				axis = -1
			)
			self._matchcost += tmp.min(axis=-1)
		self._state = AlgoState.MATCHCOST_UPDATED
		self._min_iterations = 2

	def _update_matchcost(self):
		self._matchcost = sum(
			self._metric(self._features1[i,:,:], self._features2[j,:,:])
			for i,j in self._warppath
		)
		self._state = AlgoState.MATCHCOST_UPDATED

	def _update_warpcost(self):
		self._warpcost = sum(
			self._metric(self._features1[:,v,:], self._features2[:,w,:])
			for v,w in enumerate(self._matching)
		)
		self._state = AlgoState.WARPCOST_UPDATED

	def _update_cost(self, new_cost):
		# only check after we reached a valid state
		# (initialization can set unrealistic values)
		if self.iteration > self._min_iterations and new_cost >= self._cost:
			self._state = AlgoState.CONVERGED
		self._cost = new_cost
		
	def _check_termination(self):
		if self.iteration >= self.max_iterations:
			self._state = AlgoState.ITERATION_LIMIT_REACHED
			if self.iteration < self._min_iterations:
				warning("Heuristic terminated early, result may be invalid")
		return self.terminated

	def step(self):
		if self._state is None:
			raise ValueError("Missing initialization")
		if self._check_termination():
			return
		self.iteration += 1

		if self._state == AlgoState.WARPCOST_UPDATED:
			cost, self._warppath = dynamic_timewarping(self._warpcost)
			self._update_cost(cost)
			if self._check_termination():
				return
			self._update_matchcost()

		elif self._state == AlgoState.MATCHCOST_UPDATED:
			self._matching, _, tmp = lapjv.lapjv(self._matchcost)
			self._update_cost(tmp[0])
			if self._check_termination():
				return
			self._update_warpcost()


def compute_dtgw(
	features1, features2, eps,
	metric = scipy.spatial.distance.cdist,
	observer = LoggingObserver()
):
	"""Computes the aprroximated dtgw distance between two temporal graphs.
	
	Arguments:
	----------
	features1 : numpy.ndarray of shape (l1, n1, k)
		Features of the first graph.
		features1[i,j,:] is the feature vector
		of vertex j in layer i of the first temporal graph
	features2 : numpy.ndarray of shape (l2, n2, k)
		Analogous to features1
	eps : numpy.ndarray of shape (k)
		Feature vector of a nonexisting (deleted) vertex.
	metric :
		Metric function to compute distances
	observer : 
		Observer object to notify on interal state changes

	Returns:
	---------
	d : float
		The approximate dtgw distance as determined by the alternating heuristic.
	"""

	t1, n1, k1 = features1.shape
	t2, n2, k2 = features2.shape
	n = max(n1, n2)
	features1 = np.hstack((features1, np.broadcast_to(eps, (t1, n-n1, k1))))
	features2 = np.hstack((features2, np.broadcast_to(eps, (t2, n-n2, k2))))
	heuristic = DtgwHeuristic(features1, features2, metric)
	heuristic.max_iterations = CONFIG["max iterations"]

	observer.update_features(features1=features1, features2=features2, metric=metric)

	init = CONFIG.get("heuristic initialization")
	if init == "diagonal warping":
		heuristic.initialize_with_diagonal_warping()
	elif init == "optimistic warping":
		heuristic.initialize_with_optimistic_warping()
	elif init == "optimistic matching":
		heuristic.initialize_with_optimistic_matching()
	elif init == "sigma*":
		heuristic.initiialize_with_product_warping()
	else:
		raise ValueError("Invalid initialization")

	observer.update_state("heuristic initialized",
		warp_path = heuristic.warp_path,
		matching = enumerate(heuristic.matching),
		cost = heuristic.cost,
		iteration = heuristic.iteration,
		algo_state = heuristic.state
	)

	while not heuristic.terminated:
		oldstate = heuristic.state
		heuristic.step()
		if oldstate == AlgoState.WARPCOST_UPDATED:
			observer.update_state("warp path updated", 
				warp_path = heuristic.warp_path,
				matching = enumerate(heuristic.matching),
				cost = heuristic.cost,
				iteration = heuristic.iteration,
				algo_state = heuristic.state
			)
		elif oldstate == AlgoState.MATCHCOST_UPDATED:
			observer.update_state("matching updated",
				warp_path = heuristic.warp_path,
				matching = enumerate(heuristic.matching),
				cost = heuristic.cost,
				iteration = heuristic.iteration,
				algo_state = heuristic.state
			)

	observer.update_state("done",
		iteration = heuristic.iteration,
		cost = heuristic.cost,
		matching = enumerate(heuristic.matching),
		warp_path = heuristic.warp_path,
		algo_state = heuristic.state
	)
	return heuristic.cost

